package algorithm;

import java.io.*;
import java.util.*;

import helpers.DoubleMap;
import helpers.Helpers;
import main.Finals;
import model.*;

public class EM {
	
	private static DoubleMap<Sequence> _constits;
	private static DoubleMap<Sequence> _distits;
	private static DoubleMap<Context> _constitCntxt;
	private static DoubleMap<Context> _distitCntxt;
	private static Parameters _currEst = null;
	
	public static void init()
	{
		_constits = new DoubleMap<Sequence>();
		_distits = new DoubleMap<Sequence>();
		_constitCntxt = new DoubleMap<Context>();
		_distitCntxt = new DoubleMap<Context>();
	}
	
	public static Parameters reestimate(Parameters currEst, String corpusName)
	{
		_currEst = currEst;
		init();
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(Finals.WORKING_DIR+corpusName+Finals.EXTENSION));
			String sentence = null;
			int i = 1;
			while((sentence = br.readLine())!=null)
			{
			if(!Helpers.sentenceInRange(Finals.CORPUS_SIZE/Finals.NUM_OF_LEARNERS, i, Finals.TEST_PART))
					reestimateSentence(sentence);
				if(i % 1000 == 0)
					System.out.println("reestimation: finished sentence " + i);
				i++;
			}
			br.close();
		}	
		catch(IOException e)
		{
			e.printStackTrace();
		}
		Parameters params = new Parameters(_constits, _distits, _constitCntxt, _distitCntxt);

		return params;
		
	}
	/**
	 * calculate expectation for every span according to distribution of brackets
	 * according to the current parameters
	 * @param sentence - a string that is a sentence - may contain * and #
	 */
	private static void reestimateSentence(String sentence)
	{
		Sequence sent = new Sequence();
		StringTokenizer st = new StringTokenizer(sentence);
		while(st.hasMoreTokens())
		{	
			String word = st.nextToken();
			if(!word.equals(Finals.BEGIN_SYMBOL) && !word.equals(Finals.END_SYMBOL))
				sent.add(word);
		}
		reestimateSentence(sent);	
	}
	/**
	 * estimates expectation of spans - using dynamic programming 
	 * @param sent - a vector with a sentence - no * and #
	 */
	private static void reestimateSentence(Sequence sent)
	{

		//reestimate the counts according to the curr parameters
		Bracketing insideTable = calcInsideScore(sent);
		Bracketing outsideTable = calcOutsideScore(sent,insideTable);
		double denom = insideTable.get(0,sent.size());
		
		//maximize using maximum liklihood estimate
		for(int i = 0; i < sent.size(); ++i)
		{
			for(int j = i+1; j <= sent.size(); ++j)
			{
				double expectation;
				Sequence seq = sent.getBracketingSubseq(i, j);
				Context ctxt = sent.getBracketingsContxt(i, j);
				expectation = (insideTable.get(i,j)*outsideTable.get(i,j))/denom;
				
				//sanity check
				if(expectation > 1.01 || expectation < 0)
					System.out.println("");
				
				//no need to increment zero
				if(expectation > 0)
				{
					_constits.inc(seq,expectation);
					_constitCntxt.inc(ctxt,expectation);
				}
				if(expectation < 1)
				{
					_distitCntxt.inc(ctxt,1-expectation);
					_distits.inc(seq,1-expectation);
				}
			}
		}
	}
	
	private static Bracketing calcInsideScore(Sequence sent)
	{
		Bracketing result = new Bracketing(sent);
		//the span one sequences are defined as one
		for(int i = 0; i <sent.size(); i++)
		{
			Sequence seq = sent.getBracketingSubseq(i, i+1);
			Context ctxt = sent.getBracketingsContxt(i, i+1);
			result.set(i,i+1,_currEst.phi(seq,ctxt));
		}
		//longer spans are defined dynamically
		for(int s = 2; s <= sent.size(); ++s)
		{
			for(int i = 0; i+s <= sent.size(); ++i)
			{
				int j = i+s;
				Sequence seq = sent.getBracketingSubseq(i, j);
				Context ctxt = sent.getBracketingsContxt(i, j);
				
				double sum = 0;
				for(int k = i+1; k < j; k++)
					sum += result.get(i,k)*result.get(k,j);
				
				result.set(i, j, sum*_currEst.phi(seq,ctxt));
			}
		}
		return result;
	}
	/*
	private static Bracketing calcOutsideScore2(Sequence sent, Bracketing insideTable)
	{
		Bracketing result = new Bracketing(sent);
		//for the span of the entire sentence the outside score is 1
		result.set(0, sent.size(), 1);
		//now dynamically calculate the others
		for(int s = sent.size()-1; s>0; s--)
		{
			for(int i = 0; i+s<=sent.size();++i)
			{
				int j = i+s;
				Sequence seq = sent.getBracketingSubseq(i, j);
				Context ctxt = sent.getBracketingsContxt(i, j);
				
				double sum = 0;
				for(int k = 0; k < i; k++)
					sum+=insideTable.get(k,i)*result.get(k,j);
				
				for(int k = j+1; k<=sent.size(); k++)
					sum+=insideTable.get(j,k)*result.get(i,k);
				
				result.set(i, j, sum*_currEst.phi(seq,ctxt));
			}
		}
		return result;
	} */
	private static Bracketing calcOutsideScore(Sequence sent, Bracketing insideTable)
	{
		Bracketing result = new Bracketing(sent);
		//for the span of the entire sentence the outside score is 1
		result.set(0, sent.size(), 1);
		//now dynamically calculate the others
		for(int s = sent.size()-1; s>0; s--)
		{
			for(int i = 0; i+s<=sent.size();++i)
			{
				int j = i+s;
				
				double sum = 0;
				for(int k = 0; k < i; k++)
				{
					Sequence s1 = sent.getBracketingSubseq(k, j);
					Context c1 = sent.getBracketingsContxt(k, j);
					sum+=result.get(k, j)*_currEst.phi(s1, c1)
							*insideTable.get(k, i);
				}
				for(int k = j+1; k<=sent.size();k++)
				{
					Sequence s2 = sent.getBracketingSubseq(i, k);
					Context c2 = sent.getBracketingsContxt(i, k);
					sum+=result.get(i, k)*_currEst.phi(s2, c2)
							*insideTable.get(j, k);
				}
				result.set(i, j, sum);
			}
		}
		return result;
	}
}
