package algorithm;

import helpers.DoubleMap;
import helpers.Helpers;

import java.io.*;
import java.util.*;

import model.*;
import main.Finals;

public class Initialize {
	
	private static DoubleMap<Sequence> _constits;
	private static DoubleMap<Sequence> _distits;
	private static DoubleMap<Context> _constitCntxt;
	private static DoubleMap<Context> _distitCntxt;
		
	public static void init()
	{
		_constits = new DoubleMap<Sequence>();
		_distits = new DoubleMap<Sequence>();
		_constitCntxt = new DoubleMap<Context>();
		_distitCntxt = new DoubleMap<Context>();
	}
	/**
	 * calcs the parameters empirically from the corpus - not recommended
	 * @param corpusName
	 * @return estimation of parameters
	 */
	
	public static Parameters goEmpSplit(String corpusName)
	{
		init();
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(Finals.WORKING_DIR+corpusName+Finals.EXTENSION));
			String sentence = null;
			int i = 1;
			while((sentence = br.readLine())!=null)
			{
				if(!Helpers.sentenceInRange(Finals.CORPUS_SIZE/Finals.NUM_OF_LEARNERS, i, Finals.TEST_PART))
					handleSentence(sentence);
				i++;
				if(i % 1000 == 0)
					System.out.println("Initialization: finished sentence " + i);
			}
			br.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}		
		
		Parameters params = new Parameters(_constits, _distits, _constitCntxt, _distitCntxt);
		return params;
	}
	/**
	 * takes a line that represents a sentence, creates a random tree and 
	 * increases the constituent and context counts
	 * @param sentence 
	 */
	private static void handleSentence(String sentence)
	{
		//first I need to create a tree and randomly binarize it
		TreeNode tree = new TreeNode(sentence,TreeNode.NO_BRACKETS);	
		tree.randBinarizeShallowTree();
		//TODO: then I need to use the tree to increase the counts
		countConstitsDistits(tree,sentence);		
	}
	
	private static void countConstitsDistits(TreeNode tree, String sentence)
	{
		HashSet<SubseqRange> sentConstits = tree.getConsts(0, true);
		Vector<String> sent = new Vector<String>();
		StringTokenizer st = new StringTokenizer(sentence);
		while(st.hasMoreTokens())
		{
			String word = st.nextToken();
			if(!word.equals(Finals.BEGIN_SYMBOL)&&!word.equals(Finals.END_SYMBOL))
				sent.add(word);
		}
		//first all of the one size terminals and the entire sentence are
		//constituents
		for(int i = 0; i < sent.size();++i)
		{
			Sequence seq = new Sequence(sent.get(i));
			Context ctxt = new Context(sent,new SubseqRange(i,i));
			_constits.inc(seq);
			_constitCntxt.inc(ctxt);
		}
		_constits.inc(new Sequence(sent));
		_constitCntxt.inc(new Context(Finals.BEGIN_SYMBOL,Finals.END_SYMBOL));
		//now check all of the other spans and add constits and distits
		for(int i = 0; i < sent.size()-1;++i)
		{
			for(int j = i+1; j < sent.size();++j)
			{
				SubseqRange span = new SubseqRange(i,j);
				if(sentConstits.contains(span))
				{
					_constits.inc(new Sequence(sent,span));
					_constitCntxt.inc(new Context(sent,span));
				}
				else
				{
					_distits.inc(new Sequence(sent,span));
					_distitCntxt.inc(new Context(sent,span));
				}
			}
		}
	}
	/**
	 * calcs the parameters in closed form
	 * @param corpusName
	 * @return estimation of parameters
	 */
	public static Parameters goClosedFormSplit(String corpusName)
	{
		init();
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(Finals.WORKING_DIR+corpusName+Finals.EXTENSION));
			String sentence = null;
			int i = 1;
			while((sentence = br.readLine())!=null)
			{
				if(i % 1000 == 0)
					System.out.println("Initialization: sentence " + i);
				//only train on the part of the corpus that is the train set:
				if(!Helpers.sentenceInRange(Finals.CORPUS_SIZE/Finals.NUM_OF_LEARNERS, i, Finals.TEST_PART))
					countExpectations(sentence);
				i++;
			}
			br.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}		
		
		Parameters params = new Parameters(_constits, _distits, _constitCntxt, _distitCntxt);
		return params;
	}
	/**
	 * add expectation of counts according to the closed form calculaltion
	 * from Dan klein's thesis appendix B2
	 * @param sentence
	 */
	private static void countExpectations(String sentence)
	{
		Sequence sent = new Sequence(sentence);
		//use the closed form formula
		for(int i = 0; i < sent.size();++i)
		{
			for(int j = i+1; j <= sent.size(); ++j)
			{
				double expectation;
				Sequence seq = sent.getBracketingSubseq(i, j);
				Context ctxt = sent.getBracketingsContxt(i, j);
				//the entire sentence or a one-word span
				if((i==0 && j==sent.size()) || (j-i)==1)
					expectation = 1;
				else if (i == 0 || j == sent.size())
					expectation = (double) 1/(j-i);
				else
					expectation = (double) 2 /((j-i)*(j-i+1));
										
				_constits.inc(seq,expectation);
				_constitCntxt.inc(ctxt,expectation);		
				_distits.inc(seq,1-expectation);
				_distitCntxt.inc(ctxt,1-expectation);
			}
		}	
	}
}
