package algorithm;

import model.*;
import helpers.Helpers;
import helpers.MathFuncs;

import java.io.*;
import java.util.StringTokenizer;

import probParser.*;

import main.Finals;

public class Parser {
	
	private static Parameters _params;
	
	public static void parseAndPrint(Parameters params, String corpusName)
	{
		_params = params;
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(Finals.WORKING_DIR+corpusName+Finals.EXTENSION));
			PrintWriter pw = new PrintWriter(new FileOutputStream(Finals.WORKING_DIR+Finals.GRAMMAR_NAME+".kleinParsing."+Finals.TEST_PART+Finals.EXTENSION));
			String line = null;
			int i = 1;
			while((line = br.readLine()) != null)
			{
				if(!Helpers.sentenceInRange(Finals.CORPUS_SIZE/Finals.NUM_OF_LEARNERS, i, Finals.TEST_PART))
				{
					TreeNode bestTree = parse(line);
					pw.println(bestTree.toString());
				}
				if(i % 1000 == 0)
					System.out.println("finished sentence " + i);
				i++;
			}
			br.close();
			pw.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}
	
	public static void parseAndEvaluate(Parameters params, String corpusName, String lexFile, String cfgFile)
	{
		_params = params;
		try
		{
			CFG cfg = new CFG(cfgFile,lexFile);
			CYK cyk = new CYK(cfg);
			double recallSum = 0;
			double precSum = 0;
			
			BufferedReader br = new BufferedReader(new FileReader(Finals.WORKING_DIR+corpusName+Finals.EXTENSION));
			String line = null;
			int i = 1;
			int numOfSentences = 0;
			while((line = br.readLine()) != null)
			{
				if(!Helpers.sentenceInRange(Finals.CORPUS_SIZE/Finals.NUM_OF_LEARNERS, i, Finals.TEST_PART))
				{
					TreeNode inferredTree = parse(line);
					TreeNode realTree = cyk.singleBestParse(line);
					if(realTree == null)
						System.out.println("");
					
					recallSum+=inferredTree.getUnlabeledRecall(realTree);				
					precSum+=inferredTree.getUnlabledPrecision(realTree);
					numOfSentences++;					
				}
				if(i % 1000 == 0)
					System.out.println("Parsing: finished sentence " + i);
				i++;
			}
			br.close();
			PrintWriter pw = new PrintWriter(new FileOutputStream(Finals.WORKING_DIR+"parseRes"+"."+Finals.TEST_PART+Finals.EXTENSION,true));
			double recall = (double) recallSum/numOfSentences;
			double precision = (double) precSum/numOfSentences;
			pw.println(recall+"\t"+precision+"\t"+MathFuncs.fMeasure(recall, precision));
			pw.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}
	
	public static void parseAndEvaluate(Parameters params, String corpusName, String corpusPosName, String lexFile, String cfgFile)
	{
		_params = params;
		try
		{
			CFG cfg = new CFG(cfgFile,lexFile);
			CYK cyk = new CYK(cfg);
			double recallSum = 0;
			double precSum = 0;
			
			BufferedReader br1 = new BufferedReader(new FileReader(Finals.WORKING_DIR+corpusName+Finals.EXTENSION));
			BufferedReader br2 = new BufferedReader(new FileReader(Finals.WORKING_DIR+corpusPosName+Finals.EXTENSION));
			String line1 = null;
			String line2 = null;
			int i = 1;
			int numOfSentences = 0;
			while((line1 = br1.readLine()) != null)
			{
				line2 = br2.readLine();
				if(!Helpers.sentenceInRange(Finals.CORPUS_SIZE/Finals.NUM_OF_LEARNERS, i, Finals.TEST_PART))
				{
					TreeNode inferredTree = parse(line2);
					TreeNode realTree = cyk.singleBestParse(line1);
					if(realTree == null)
						System.out.println("");
					
					if (realTree.getLength() > 2)
					{
						recallSum+=inferredTree.getUnlabeledRecall(realTree);				
						precSum+=inferredTree.getUnlabledPrecision(realTree);
						numOfSentences++;
					}
					
				}
				if(i % 1000 == 0)
					System.out.println("Parsing: finished sentence " + i);
				i++;
			}
			br1.close();
			br2.close();
			PrintWriter pw = new PrintWriter(new FileOutputStream(Finals.WORKING_DIR+"parseRes"+"."+Finals.TEST_PART+Finals.EXTENSION,true));
			double recall = (double) recallSum/numOfSentences;
			double precision = (double) precSum/numOfSentences;
			pw.println(recall+"\t"+precision+"\t"+MathFuncs.fMeasure(recall, precision));
			pw.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}
	
	public static void parseAndEvaluate(Parameters params, String corpusName, String bracketedCorpusName)
	{
		_params = params;
		try
		{
			BufferedReader brCorpus = new BufferedReader(new FileReader(Finals.WORKING_DIR+corpusName+Finals.EXTENSION));
			BufferedReader brBracketed = new BufferedReader(new FileReader(Finals.WORKING_DIR+bracketedCorpusName+Finals.EXTENSION));
			String corpusLine = null;
			String bracketedLine = null;
			int i = 1;
			int numOfSentences = 0;
			double recallSum=0;
			double precSum=0;
			while((corpusLine = brCorpus.readLine()) != null)
			{
				bracketedLine = brBracketed.readLine();
				if(!Helpers.sentenceInRange(Finals.CORPUS_SIZE/Finals.NUM_OF_LEARNERS, i, Finals.TEST_PART))
				{
					TreeNode inferredTree = parse(corpusLine);
					TreeNode realTree = new TreeNode(bracketedLine,TreeNode.BRACKETS);
					
					recallSum+=inferredTree.getUnlabeledRecall(realTree);				
					precSum+=inferredTree.getUnlabledPrecision(realTree);
					numOfSentences++;
				}
				if(i % 1000 == 0)
					System.out.println("Parsing: finished sentence " + i);
				i++;
			}
			brBracketed.close();
			brCorpus.close();
			PrintWriter pw = new PrintWriter(new FileOutputStream(Finals.WORKING_DIR+"parseRes"+"."+Finals.TEST_PART+Finals.EXTENSION,true));
			double recall = (double) recallSum/numOfSentences;
			double precision = (double) precSum/numOfSentences;
			pw.println(recall+"\t"+precision+"\t"+MathFuncs.fMeasure(recall, precision));
			pw.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
		
	}
	
	private static TreeNode parse(String sentence)
	{
		Sequence sent = new Sequence();
		StringTokenizer st = new StringTokenizer(sentence);
		while(st.hasMoreTokens())
		{	
			String word = st.nextToken();
			if(!word.equals(Finals.BEGIN_SYMBOL) && !word.equals(Finals.END_SYMBOL))
				sent.add(word);
		}
		return parse(sent);	
	}
	
	private static TreeNode parse(Sequence sent)
	{
		ParseTable table = new ParseTable(sent);
		//first let's fill the one-span cells in the table
		for(int i = 0; i < sent.size(); ++i)
		{
			Sequence seq = sent.getBracketingSubseq(i, i+1);
			Context ctxt = sent.getBracketingsContxt(i, i+1);
			double score = _params.phi(seq, ctxt);
			TreeNode termTree = new TreeNode(seq.get(0),TreeNode.LABEL,score);
			TreeNode preTerm = new TreeNode(Finals.PRE_TERMINAL,TreeNode.LABEL,score);
			preTerm.addChild(termTree);			
			table.set(i, i+1, preTerm);
		}
		
		for(int span = 2; span <= sent.size(); ++span)
		{
			for(int i = 0; i+span<=sent.size(); i++)
			{
				int j = i+span;
				Sequence seq = sent.getBracketingSubseq(i, j);
				Context ctxt = sent.getBracketingsContxt(i, j);
				double rootScore = _params.phi(seq, ctxt);
				for(int k = i+1; k < j; ++k)
				{
					double newScore = 
						table.get(i, k).getProb()* 
						table.get(k, j).getProb()*
						rootScore;
					
					if(table.get(i, j)== null || table.get(i, j).getProb() < newScore)
					{
						TreeNode newTree = new TreeNode(Finals.NON_TERMINAL,TreeNode.LABEL,newScore);
						newTree.addChild(table.get(i, k));
						newTree.addChild(table.get(k, j));
						table.set(i, j, newTree);
					}
					
				}
			}
		}	
		return table.get(0, sent.size());
	}

}
