package main;

import java.util.Date;

import model.*;
import algorithm.*;

public class Main {
	
	//the part of the corpus to use as a test set
	public static int learner = Finals.TEST_PART;
	public static String action = "";
	
	public static void main(String args[]) {
		
		Date startTime = new Date();
		
		if (args.length%2==1)
		{
			System.out.println("Illegal number of arguments");
			return;
		}
			
		for (int i=0;i<args.length;i+=2) 
		{
			char flag = (char)args[i].charAt(1);
			String value = args[i+1];
			
			switch (flag)
			{
				//choosing the part of the corpus to test on
				case 'f':
					Finals.CORPUS_NAME = value;
					break;
				case 'b':
					Finals.BRACKETED_CORPUS_NAME = value;
					break;
				case 'w':
					Finals.WORKING_DIR = value;
					break;
				case 'e':
					Finals.EXTENSION = value;
					break;
				case 'r':
					Finals.ROOT = value;
					break;
				case 't':
					Finals.TEST_PART = Integer.parseInt(value);
					break;
				case 's':
					Finals.CORPUS_SIZE = Integer.parseInt(value);
					break;
				case 'l':
					Finals.LEXICON = value;
					break;
				case 'g':
					Finals.GRAMMAR = value;
					break;
				case 'I':
					Finals.NUM_OF_ITERATIONS = Integer.parseInt(value);
					break;
				case 'L':
					Finals.NUM_OF_LEARNERS = Integer.parseInt(value);
					break;
				default:
					System.out.println("Unknown flag -"+flag);
					break;
			}
		}
				
		train(Finals.CORPUS_NAME);
		
		Date endTime = new Date();
		long runTime = endTime.getTime()-startTime.getTime();
		System.out.println("Runtime (in minutes): " + ((double)(runTime/1000))/60);
	}
	
	public static void train(String corpusName)
	{
		//initialization of parameters
		Parameters params = Initialize.goClosedFormSplit(Finals.CORPUS_NAME);
		if(!Finals.BRACKETED_CORPUS_NAME.equals(""))
				Parser.parseAndEvaluate(params, Finals.CORPUS_NAME,Finals.BRACKETED_CORPUS_NAME);
		for(int i = 1; i <= Finals.NUM_OF_ITERATIONS; ++i)
		{
			params = EM.reestimate(params, corpusName);
			if(!Finals.BRACKETED_CORPUS_NAME.equals(""))
				Parser.parseAndEvaluate(params, Finals.CORPUS_NAME,Finals.BRACKETED_CORPUS_NAME);
			System.out.println("Finished iteration: " + i);
		}
		Parser.parseAndPrint(params, corpusName);
	}
}
