package model;

import algorithm.Smoother;
import helpers.DoubleMap;
public class Parameters {
	
	private DoubleMap<Sequence> _constits;
	private DoubleMap<Sequence> _distits;
	private DoubleMap<Context> _constitCntxt;
	private DoubleMap<Context> _distitCntxt;
	
	public Parameters(DoubleMap<Sequence> constits, DoubleMap<Sequence> distits,
			DoubleMap<Context> constitCntxt, DoubleMap<Context> distitCntxt)
	{
		_constits = constits;
		_distits = distits;
		_constitCntxt = constitCntxt;
		_distitCntxt = distitCntxt;
	}
	
	public DoubleMap<Sequence> getConstits()
	{
		return _constits;
	}
	
	public DoubleMap<Sequence> getDistits()
	{
		return _distits;
	}
	
	public DoubleMap<Context> getConstitCntxt()
	{
		return _constitCntxt;
	}
	
	public DoubleMap<Context> getDistitCntxt()
	{
		return _distitCntxt;
	}
	
	public double phi(Sequence seq, Context ctxt)
	{
		double seqConstitCount = _constits.get(seq) + Smoother.CONSTIT_SMOOTHER;
		double seqConstitTotal = _constits.getTotal()+Smoother.CONSTIT_SMOOTHER*_constits.size();
		double seqConstitProb = seqConstitCount/seqConstitTotal; 
		
		double ctxtConstitCount = _constitCntxt.get(ctxt) + Smoother.CONSTIT_SMOOTHER;
		double ctxtConstitTotal = _constitCntxt.getTotal()+Smoother.CONSTIT_SMOOTHER*_constitCntxt.size();
		double ctxtConstitProb = ctxtConstitCount/ctxtConstitTotal; 
			
		double seqDistitCount = _distits.get(seq) + Smoother.DISTIT_SMOOTHER;
		double seqDistitTotal = _distits.getTotal()+Smoother.DISTIT_SMOOTHER*_distits.size();
		double seqDistitProb = seqDistitCount/seqDistitTotal; 
		
		double ctxtDistitCount = _distitCntxt.get(ctxt) + Smoother.DISTIT_SMOOTHER;
		double ctxtDistitTotal = _distitCntxt.getTotal()+Smoother.DISTIT_SMOOTHER*+_distitCntxt.size();
		double ctxtDistitProb = ctxtDistitCount/ctxtDistitTotal; 
		
		return (seqConstitProb*ctxtConstitProb) / 
				(seqDistitProb*ctxtDistitProb);
	}
	
}
