package model;

import java.util.StringTokenizer;
import java.util.Vector;
import main.*;

public class Sequence extends Vector<String> implements Comparable{
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	public Sequence()
	{
		super();
	}
	
	public Sequence(String seq)
	{
		super();
		StringTokenizer st = new StringTokenizer(seq);
		while(st.hasMoreTokens())
		{
			String word = st.nextToken();
			if(!word.equals(Finals.BEGIN_SYMBOL) && !word.equals(Finals.END_SYMBOL))
				add(word);
		}
	}
	
	public Sequence(Vector<String> vec)
	{
		super();
		for(String str: vec)
			add(str);
	}
	/**
	 * @param vec vector of words
	 * @param subseq the range to extract from the 'from' until the 'to' inclusive
	 */
	public Sequence(Vector<String> vec, SubseqRange subseq)
	{
		super();
		for(int i = subseq.getFrom(); i <= subseq.getTo(); ++i)
			add(vec.get(i));
	}
	/**
	 * get a sequence according to the bracketing indexing
	 * @param from - index i is after between word i and i+1
	 * @param to
	 * @return
	 */
	public Sequence getBracketingSubseq(int from, int to)
	{
		Sequence result = new Sequence();
		for(int i = from; i < to; ++i)
			result.add(get(i));
		return result;
	}
	
	/**
	 * we create a context exclusively, i.e. without the word in index 'to'
	 * @param from
	 * @param to
	 * @return
	 */
	public Context getBracketingsContxt(int from, int to)
	{
		Context result = new Context("","");
		if(from==0)
			result.setLeft("*");
		else
			result.setLeft(get(from-1));
		if(to==size())
			result.setRight("#");
		else
			result.setRight(get(to));
		return result;
	}
	
	public boolean equals(Object o)
	{
		if(!(o instanceof Vector))
			return false;
		Vector other = (Vector) o;
		if(other.size()!=size())
			return false;
		else
			for(int i = 0; i<size();++i)
				if(!get(i).equals(other.get(i)))
					return false;
		return true;
		
	}
	public int hashCode()
	{
		return toString().hashCode();
	}
	
	public int compareTo(Object o) {
		return toString().compareTo(o.toString());
	}
}
