package model;

import helpers.DoubleMap;
import helpers.Helpers;

import java.util.*;
import java.io.*;

import probParser.Rule;
import main.Finals;

public class TreeNode {
	
	String _label;
	double _probability;
	Vector<TreeNode> _children;
	
	public final static int LABEL=0;
	public final static int BRACKETS = 1;
	public final static int NO_BRACKETS = 2;
	
	public TreeNode(DoubleMap<String>[][] probChart, HashMap<String,Rule>[][] ruleChart, HashMap<String,Integer>[][] indexChart, int i, int j, String label)
	{
		_label = label;
		_children = new Vector<TreeNode>();
		if(i == 0) // it is a preterminal rule this is our stopping condition
		{
			Rule rule = ruleChart[i][j].get(label);
			_children.add(new TreeNode(rule.getRHS().get(0),TreeNode.LABEL));
		}
		else
		{
			_probability = probChart[i][j].get(label);
			String leftLabel = ruleChart[i][j].get(label).getRHS().get(0);
			String rightLabel = ruleChart[i][j].get(label).getRHS().get(1);
			int k = indexChart[i][j].get(label).intValue();
			TreeNode leftChild = new TreeNode(probChart,ruleChart,indexChart,k,j,leftLabel);
			TreeNode rightChild = new TreeNode(probChart,ruleChart,indexChart,i-1-k,j+k+1,rightLabel);
			_children.add(leftChild);
			_children.add(rightChild);
		}	
	}
		
	public TreeNode(String str, int type)
	{
		if(type==LABEL)
		{
			_label = str;
			_probability = 1.0;
			_children = new Vector<TreeNode>();
		}
		//this creates a shaloow tree from a string
		else if(type==NO_BRACKETS)
		{
			_label = Finals.ROOT;
			_children = new Vector<TreeNode>();
			StringTokenizer st = new StringTokenizer(str);
			while(st.hasMoreTokens())
			{
				String word = st.nextToken();
				if(!word.equals(Finals.BEGIN_SYMBOL) && !word.equals(Finals.END_SYMBOL))
					_children.add(new TreeNode(word,LABEL));
			}
		}
		else if(type==BRACKETS) //takes a bracketed string and parses it
		{
			_children = new Vector<TreeNode>();
			_probability = 1.0;
			//right now format means nothing
			if(!str.startsWith("("))
				_label = str;
			else //recursive case
			{
				//first strip parentheses
				int lastClosingBracket = str.lastIndexOf(')');
				String strippedStr = str.substring(1,lastClosingBracket);
				//check if it's a root
				if(Helpers.getFirstNonWhitespaceChar(strippedStr)== '(')
				{
					_label = "ROOT_0";
					strippedStr = strippedStr.substring(strippedStr.indexOf('('));
					_children.add(new TreeNode(strippedStr,BRACKETS));
				}
				else
				{
					StringTokenizer st = new StringTokenizer(strippedStr);
					_label = st.nextToken();
					//now add all the children
					//rid the label
					String isTerminal = st.nextToken();
					if(!isTerminal.startsWith("("))
						_children.add(new TreeNode(isTerminal,0));
					else
					{
						while(Helpers.getFirstNonWhitespaceChar(strippedStr) != ' ')
						{
							strippedStr = strippedStr.substring(strippedStr.indexOf('('));
							int closingBracketInd = Helpers.findMatchingBracket(strippedStr,0);
							_children.add(new TreeNode(strippedStr.substring(0,closingBracketInd+1),BRACKETS));
							strippedStr = strippedStr.substring(closingBracketInd+1);
						}
					}
				}
			}				
		}
	}
	
	public TreeNode(String label, Vector<TreeNode> children)
	{
		_label = label;
		_probability = 1.0;
		_children = children;
	}
	
	public TreeNode(String str, int type, double probability)
	{
		if(type==LABEL)
		{
			_label = str;
			_probability = probability;
			_children = new Vector<TreeNode>();
		}
		//this creates a shaloow tree from a string
		else if(type==NO_BRACKETS)
		{
			_label = Finals.ROOT;
			_children = new Vector<TreeNode>();
			StringTokenizer st = new StringTokenizer(str);
			while(st.hasMoreTokens())
			{
				String word = st.nextToken();
				if(!word.equals(Finals.BEGIN_SYMBOL) && !word.equals(Finals.END_SYMBOL))
				
					_children.add(new TreeNode(word,LABEL));
			}
			_probability = probability;
		}
	}
	
	/**
	 * condition: every sentence starts with a "("
	 * after an '(' the non terminal appears with no whitespace in between
	 * no \r or \n in the string
	 * @param pennString
	 * @param format
	 */
	
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		if(_children.size() == 0)
			return _label;
		if(_children.size()==1)
		{
			sb.append("(" + _label + " " + _children.get(0).toString() + ")");
			return sb.toString();
		}
		else
		{
			sb.append("(" + _label +  "  ");
			for(TreeNode child: _children)
				sb.append(child.toString()+ " ");
				sb.append(")");
			return sb.toString();
		}
	}
	
	public String toPennString()
	{
		return "(" + toString() + ")";
	}
	public double getProb()
	{
		return _probability;
	}
	/**	
	 * @return length of the yield
	 */
	public int getLength()
	{
		int sum = 0;
		if(_children.size()==0)
			return 1;
		if(_children.size()>0)
		{
			for(TreeNode child: _children)
				sum+=child.getLength();
		}
		return sum;
	}
	
	public String getYield()
	{
		if(_children.size()==0)
			return _label;
		if(_children.size()==1)
		{
			StringBuffer sb = new StringBuffer();
			sb.append(_children.get(0).getYield());
			return (sb.toString());	
		}
		else
		{
			StringBuffer sb = new StringBuffer();
			for(TreeNode child: _children)
				sb.append(child.getYield() + " ");
			return (sb.toString());
		}
	}
	
	public String getLabel()
	{
		return _label;
	}
	public void addChild(TreeNode child)
	{
		_children.add(child);
	}
	public TreeNode getChild(int i)
	{
		return _children.get(i);
	}
	
	/**
	 * uncnfize the longer than 2 rules but doesn't reinstate the unary rules lost
	 * @return
	 */
	/*
	private TreeNode partialUnCnfize()
	{
		if (_children.size()==0)
			return new TreeNode(_label,LABEL);
		else
		{
			TreeNode resultTree = new TreeNode(_label,LABEL);
			for (TreeNode child: _children)
			{
				if(!child._label.contains("["))
					resultTree.addChild(child.partialUnCnfize());
				else
				{				
					if(child.getLabel().equals("[s]"))
						System.out.print("");
					Vector<TreeNode> flatChildren = child.flattenCNFizedRule();
					for(int i = 0; i < flatChildren.size();++i)
						flatChildren.set(i, flatChildren.get(i).partialUnCnfize());
					
					resultTree._children.addAll(flatChildren);
				}
			}
			return resultTree;	
		}
	}
	
	private Vector<TreeNode> flattenCNFizedRule()
	{
		Vector<TreeNode> result = new Vector<TreeNode>();
		if(!_label.contains(("[")))
		{
			result.add(this);
			return result;
		}
		else
		{
			
			if(_children.size()==1)
			{
				result.add(getChild(0));
				return result;
			}
				
			result.addAll(_children.get(0).flattenCNFizedRule());
			
			if(_children.get(1)._label.contains("["))
				result.addAll(_children.get(1).flattenCNFizedRule());	
			else 	
				result.add(_children.get(1));
			
			return result;
		}				
	}
	*/
	
	/**
	 * 
	 * @return the constituents that are not the whole sentence or of size 1
	 */
	public HashSet<SubseqRange> getConsts(int beginIndex,boolean isRoot)
	{
		HashSet<SubseqRange> result = new HashSet<SubseqRange>();
		if (_children.size()==0)
			return result;
		else if(_children.size()==1)
		{
			result.addAll(_children.get(0).getConsts(beginIndex, isRoot));
			return result;
		}
		else
		{
			int currBeginIndex = beginIndex;
			for(TreeNode child : _children)
			{
				result.addAll(child.getConsts(currBeginIndex,false));
				currBeginIndex += child.getLength();
			}
			if(!isRoot && getLength()>1)
				result.add(new SubseqRange(beginIndex,beginIndex+getLength()-1));	
			return result;
		}
	}
	//this filters out consts that are there due to CNFization
	public HashSet<SubseqRange> getRealConsts(int beginIndex,boolean isRoot)
	{
		HashSet<SubseqRange> result = new HashSet<SubseqRange>();
		if (_children.size()==0)
			return result;
		else if(_children.size()==1)
		{
			result.addAll(_children.get(0).getRealConsts(beginIndex, isRoot));
			return result;
		}
		else
		{
			int currBeginIndex = beginIndex;
			for(TreeNode child : _children)
			{
				result.addAll(child.getRealConsts(currBeginIndex,false));
				currBeginIndex += child.getLength();
			}
			if(!isRoot && getLength()>1 && !_label.startsWith("["))
				result.add(new SubseqRange(beginIndex,beginIndex+getLength()-1));	
			return result;
		}
	}
	 
	/**
	 * 
	 * @param beginIndex
	 * @param isRoot
	 * @return all constituents including one word and entire sentence
	 */
	public HashSet<SubseqRange> getAllConsts(int beginIndex)
	{
		HashSet<SubseqRange> result = new HashSet<SubseqRange>();
		if (_children.size()==0)
		{
			result.add(new SubseqRange(beginIndex,beginIndex));
			return result;
		}
		else if(_children.size()==1)
		{
			result.addAll(_children.get(0).getAllConsts(beginIndex));
			return result;
		}
		else
		{
			int currBeginIndex = beginIndex;
			for(TreeNode child : _children)
			{
				result.addAll(child.getAllConsts(currBeginIndex));
				currBeginIndex += child.getLength();
			}
			result.add(new SubseqRange(beginIndex,beginIndex+getLength()-1));	
			return result;
		}
	}
	
	public Vector<String> getConstitYields()
	{
		Vector<String> result = new Vector<String>();
		if(_children.size()==0)
			result.add(getYield());
		else
		{
			result.add(getYield());
			for(TreeNode child: _children)
				result.addAll(child.getConstitYields());
			
		}
		return result;
	}
	
	public int height()
	{
		if(_children.size()==0)
			return 1;
		else
		{
			int max = 1;
			for(TreeNode child: _children)
			{
				if(child.height() > max)
					max = child.height();
			}
			return max+1;
		}
	}
	
	/**
	 * takes a shallow tree and binarizes it recursively
	 * @return the same object with a random binary tree
	 */
	public void randBinarizeShallowTree()
	{
		//stop condition
		if(_children.size()< 3)
			return;
			
		//for a sentence with n words there are n-1 possible splits
		Random generator = new Random();
		int splitPoint = generator.nextInt(_children.size()-1);
		//create two children
		TreeNode left = new TreeNode("CAT",LABEL);
		TreeNode right = new TreeNode("CAT",LABEL);
		for(int i = 0; i< _children.size();++i)
		{
			if(i<=splitPoint)
				left.addChild(_children.get(i));
			else
				right.addChild(_children.get(i));
		}
		//recursive calls
		left.randBinarizeShallowTree();
		right.randBinarizeShallowTree();
		//changing the object
		_children.clear();
		_children.add(left);
		_children.add(right);
	}
	
	public double getUnlabeledRecall(TreeNode realTree)
	{
		HashSet<SubseqRange> inferConsts = getRealConsts(0, true);
		HashSet<SubseqRange> realConsts = realTree.getRealConsts(0, true);
		
		int realSize = realConsts.size();
		//if there are no constituents then we have succeeded
		if(realSize==0)
			return 1;
		inferConsts.retainAll(realConsts);
		return (double) inferConsts.size()/realSize;		
	}
	public double getUnlabledPrecision(TreeNode realTree)
	{
		HashSet<SubseqRange> inferConsts = getRealConsts(0, true);
		HashSet<SubseqRange> realConsts = realTree.getRealConsts(0, true);
		
		int inferSize = inferConsts.size();
		if(inferSize==0)
			return 1;
		
		inferConsts.retainAll(realConsts);
		return (double) inferConsts.size()/inferSize;
	}
	
	public static void printTrees(Vector<TreeNode> trees)
	{
		try
		{
			PrintWriter pw = new PrintWriter(new FileOutputStream(Finals.WORKING_DIR+"wsj"+Finals.EXTENSION));
			for(TreeNode tree: trees)
				pw.println(tree.toString());
			pw.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}
	
	public static void printPoss(Vector<TreeNode> trees)
	{
		try
		{
			PrintWriter pw = new PrintWriter(new FileOutputStream(Finals.WORKING_DIR+"pos"+Finals.EXTENSION));
			for(TreeNode tree: trees)
				pw.println("* "+tree.getPosYield()+" #");
			pw.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
		
	}
	
	/**
	 * It is guaranteed that a sentence contains at least one non-empty and non
	 * punctation element
	 * @return pruned tree
	 */
	public TreeNode prunePunctationAndEmpties()
	{
		Vector<TreeNode> result = new Vector<TreeNode>();
		if(isPuncOrEmpty())
			return null;
		else if(_children.size()==0)
			return new TreeNode(_label,0);
		else
		{
		for(TreeNode child: _children)
		{
			TreeNode newChild = child.prunePunctationAndEmpties();
			if(newChild != null)
				result.add(newChild);
		}
		if(result.size()>0)
			return new TreeNode(_label,result);
		return null;
		}
	}
	
	public boolean isPuncOrEmpty()
	{
		if( _label.equals(",") ||
			_label.equals(".") ||
			_label.equals(":") ||
			_label.equals("``") ||
			_label.equals("''") ||
			_label.equals("$") ||
			_label.equals("#") ||
			_label.equals("-LRB-") ||
			_label.equals("-RRB-") ||
			_label.equals("-NONE-"))
			return true;
		return false;
	}
	
	public String getPosYield()
	{
		if(_children.size()==1)
		{
			if(_children.get(0)._children.size()==0)
				return _label;
			else
			{
				StringBuffer sb = new StringBuffer();
				sb.append(_children.get(0).getPosYield());
				return (sb.toString());
			}
		}
		else
		{
			StringBuffer sb = new StringBuffer();
			for(TreeNode child: _children)
				sb.append(child.getPosYield() + " ");
			return (sb.toString());
		}
		
	}
	
}

