package probParser;

import java.io.*;
import java.util.*;

import main.Finals;
import helpers.MultiMap;
import helpers.DoubleMap;

// A CNF CFG. Can load any sort of grammar, but transforms it on loading
public class CFG 
{
	public Vector<Rule> rules;
	public MultiMap<String, Rule> rulesByLhs;
	public HashSet<String> nonTerminals;
	
	public MultiMap<String, Rule> rulesByTerminalRHS;
	public HashMap<String, MultiMap<String, Rule>> rulesByRHS0;
	private String _startSymbol = "SIGMA";
	
	
		
	public CFG(String filename) {
		rules = new Vector<Rule>();
		rulesByLhs = new MultiMap<String, Rule>();
		nonTerminals = new HashSet<String>();
		HashMap<String, Rule> stringToRule = new HashMap<String, Rule>();
		
		try {
			BufferedReader br = new BufferedReader(new FileReader(Finals.WORKING_DIR + filename));
			String line;
			
			while ((line = br.readLine()) != null) {
				Rule rule = new Rule(line);
				nonTerminals.add(rule.getLHS());
				
				if (stringToRule.containsKey(rule.toShortString())) {
					stringToRule.get(rule.toShortString()).prob += rule.prob;
				} else {
					rules.add(rule);
					rulesByLhs.addValue(rule.getLHS(), rule);
					stringToRule.put(rule.toShortString(), rule);
				}
			}
			
			br.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		System.out.println("CNFying...");
		makeCNF();
		initNonTerminals();
		finalizeNonTerminals();
	}
	
	public CFG(String grammarFile, String lexFile) {
		rules = new Vector<Rule>();
		rulesByLhs = new MultiMap<String, Rule>();
		
		try {
		    BufferedReader br = new BufferedReader(new FileReader(Finals.WORKING_DIR + lexFile));
		    String line;
		    
		    while ((line = br.readLine()) != null) {
		        StringTokenizer st = new StringTokenizer(line, " ");
		        String left, right;
		        left = st.nextToken();
		        right = st.nextToken();
		        double probability = 0.0;
		        if(st.hasMoreTokens())
		        	probability = (new Double(st.nextToken())).doubleValue();
		        
		        if (!left.equals(right)) {
		            Vector<String> v = new Vector<String>();
		            v.add(left);
		            Rule newRule = new Rule(right, v,probability);
		            rules.add(newRule);
		            rulesByLhs.addValue(right, newRule);
		        }   	        
		    }
		    
		    br.close();
		    br = new BufferedReader(new FileReader(Finals.WORKING_DIR + grammarFile));
		    
		    while ((line = br.readLine()) != null) {
		        if (line.length() == 0) continue;
		        
		        String left = line;
		        while ((line = br.readLine()) != null && line.length() != 0) {
		            StringTokenizer st = new StringTokenizer(line, " ");
		            Vector<String> v = new Vector<String>();
		            double probability = 0.0;
		            while (st.hasMoreTokens()) 
		            {
		            	String token = st.nextToken();
		            	char firstChar = token.charAt(0);
		            	if(firstChar >= '0' && firstChar <= '9')
		            		probability = (new Double(token)).doubleValue();
		            	else
		            		v.add(token);
		            }
		            
		            Rule newRule = new Rule(left, v, probability);
		            rules.add(newRule);
		            rulesByLhs.addValue(left, newRule);
		        }
		    }
		} catch (Exception e) {
		    e.printStackTrace();
		}
		
		initNonTerminals();
		finalizeRuleProbabilities();
		System.out.println("CNFying...");
		makeCNF();
		initNonTerminals();
		finalizeNonTerminals();
	}
		
	private void makeCNF() {
		// First transform every rule that transforms a non-terminal to
		// a non-terminal into a bunch of auxilliary rules
		HashMap<String, Rule> hsRules = new HashMap<String, Rule>();
		for (Rule rule : rules)
			hsRules.put(rule.toShortString(), rule);
		
		int i = 0;
		while (i < rules.size()) {
			System.out.println("First transformation - " + rules.size() + " rules...");
			Rule rule = rules.get(i);

			if (rule.RHS.size() == 1 && isNonTerminal(rule.RHS.get(0))) {
				for (int j=0; j<rules.size(); j++) {
					Rule rule2 = rules.get(j);
					if (rule2.LHS.equals(rule.RHS.get(0))) {
						Rule newRule = new Rule(rule.LHS, rule2.RHS, rule.prob*rule2.prob);
						
						if (!hsRules.containsKey(newRule.toShortString())) {
							rules.add(newRule);
							hsRules.put(newRule.toShortString(), newRule);
						} //else {
						//	hsRules.get(newRule.toShortString()).prob += newRule.prob;
						//}
					}
				}
				
				rules.remove(i);
				hsRules.remove(rule.toShortString());
			} else
				i++;
		}
		
		ensureProbabilities();
		
		// Complete the transformation
		HashSet<String> coveredTerminals = new HashSet<String>();
		for (i=0; i < rules.size(); i++) {
			System.out.println("Second transformation - " + rules.size() + " rules...");
			Rule rule = rules.get(i);
			
			if (rule.RHS.size() == 1) // necessarily a terminal
				continue;
			
			if(rule.RHS.size()==0)
				System.out.print("");
			
			if (!isNonTerminal(rule.RHS.get(0))) 
				rule.RHS.set(0, newNonTerminal(rule.RHS.get(0), coveredTerminals));
			
			if (rule.RHS.size() == 2) {
				if (!isNonTerminal(rule.RHS.get(1)))
					rule.RHS.set(1, newNonTerminal(rule.RHS.get(1), coveredTerminals));
				
				continue;
			}
			
			// necessary to split rule
			rule.RHS.set(1, newNonTerminal(rule.RHS, coveredTerminals));
			while (rule.RHS.size() > 2)
				rule.RHS.remove(2);
		}

		ensureProbabilities();
	}
	
	private String newNonTerminal(String RHS, HashSet<String> coveredTerminals) {
		String name = "[" + RHS + "]";
		
		if (coveredTerminals.contains(RHS))
			return name;
		
		Vector<String> newRHS = new Vector<String>();
		newRHS.add(RHS);
		Rule newRule = new Rule(name, newRHS, 1.0);
		rules.add(newRule);
		coveredTerminals.add(RHS);
		nonTerminals.add(name);
		
		return name;
	}
	
	private String newNonTerminal(Vector<String> unsplitRHS, HashSet<String> coveredTerminals) {
		Vector<String> newRHS = new Vector<String>(unsplitRHS);
		newRHS.remove(0);
		String name = "[" + newRHS.toString() + "]";
		if (coveredTerminals.contains(newRHS.toString()))
			return name;

		Rule newRule = new Rule(name, newRHS, 1.0);
		rules.add(newRule);
		coveredTerminals.add(newRHS.toString());
		nonTerminals.add(name);
		
		return name;
	}
	
	private void ensureProbabilities() {
		DoubleMap<String> probs = new DoubleMap<String>();
		for (Rule r : rules) {
			probs.inc(r.LHS, r.prob);
		}
		
		for (String lhs : probs.keySet())
			if (Math.abs(1 - probs.get(lhs)) > 0.00001)
				System.out.print("");
	}
	
	/*
	private void leftFactor(String symbol, Vector<Vector<String>> rhss) {		
		if (rhss.size() == 0)
			return;
		
		if (rhss.size() == 1) {
			Rule r = new Rule(symbol, rhss.get(0));
			rules.add(r);
			rulesByLhs.addValue(symbol, r);
			return;
		}
		
		Collections.sort(rhss, new StringVectorComparator());
		boolean empty = false;
		while (rhss.size() > 0) {
			lastNonTerminal++;
			Vector<Vector<String>> rhss2 = new Vector<Vector<String>>();
			Vector<String> rhs = rhss.remove(0);
			String first = rhs.remove(0);
			if (rhs.size() > 0)
				rhss2.add(rhs);
			else
				empty = true;
			
			Iterator<Vector<String>> it = rhss.iterator();
			while (it.hasNext()) {
				Vector<String> rhs2 = it.next();
				if (rhs2.get(0).equals(first)) {
					it.remove();
					if (rhs2.size() > 1) {
						rhs2.remove(0);
						rhss2.add(rhs2);
					} else
						empty = true;
				}
			}
			
			Vector<String> factoredRHS = new Vector<String>();
			factoredRHS.add(first);
			factoredRHS.add("_LeftFactorization" + lastNonTerminal);
			Rule rule = new Rule(symbol, factoredRHS);
			rules.add(rule);
			rulesByLhs.addValue(symbol, rule);
			
			if (empty) {
				Vector<String> rhs2 = new Vector<String>();
				rhs2.add(first);
				Rule rule2 = new Rule(symbol, rhs2);
				rules.add(rule2);
				rulesByLhs.addValue(symbol, rule2);
			}
			
			leftFactor("_LeftFactorization" + lastNonTerminal, rhss2);
		}
	}
	*/	
		
	private void initNonTerminals() {
		nonTerminals = new HashSet<String>();
		for (int i=0; i<rules.size(); i++) {
			String nonTerminal = rules.get(i).getLHS();
			if (!nonTerminals.contains(nonTerminal))
				nonTerminals.add(nonTerminal);
		}
	}
	
	private void finalizeNonTerminals() {
		rulesByTerminalRHS = new MultiMap<String, Rule>();
		rulesByRHS0 = new HashMap<String, MultiMap<String, Rule>>();
		
		for (Rule rule : rules) {
			if (rule.RHS.size() == 1)
				rulesByTerminalRHS.addValue(rule.RHS.get(0), rule);
			else {
				String rhs0 = rule.RHS.get(0);
				String rhs1 = rule.RHS.get(1);
				MultiMap<String, Rule> mm;
				
				if (!rulesByRHS0.containsKey(rhs0)) {
					mm = new MultiMap<String, Rule>();
					rulesByRHS0.put(rhs0, mm);
				} else
					mm = rulesByRHS0.get(rhs0);
				
				mm.addValue(rhs1, rule);
			}
		}
	}
	
	public HashSet<Rule> getRules(String rhs0, String rhs1) {
		if (!rulesByRHS0.containsKey(rhs0))
			return null;
		
		MultiMap<String, Rule> mm = rulesByRHS0.get(rhs0);
		if (!mm.containsKey(rhs1))
			return null;
		
		return mm.getAll(rhs1);
	}
	
	public void limitRecursionDepth(int depth) {
	
		rules = new Vector<Rule>();
		MultiMap<String, Rule> oldRulesByLhs = rulesByLhs;
		rulesByLhs = new MultiMap<String, Rule>();
		HashSet<String> newNonTerminals = new HashSet<String>();
			
		Vector<String> nextSymbols = new Vector<String>();
		nextSymbols.add(_startSymbol);
		int level = 0;
		
		while (nextSymbols.size() > 0) {
			Vector<String> nextNextSymbols = new Vector<String>();
			for (String curSymbol : nextSymbols) {
				String newName;
				if (curSymbol.equals(_startSymbol))
					newName = _startSymbol;
				else
					newName = curSymbol + level;
				
				for (Rule rule : oldRulesByLhs.getAll(curSymbol)) {
					if (rule.hasNonTerminals() && level == depth) 
						continue;
					else {
						for (String nextNextSymbol : rule.nonTerminals())
							if (!nextNextSymbols.contains(nextNextSymbol))
								nextNextSymbols.add(nextNextSymbol);
						
						Vector<String> newRHS = new Vector<String>();
						for (String s : rule.getRHS()) {
							if (isNonTerminal(s))
								newRHS.add(s + (level+1));
							else
								newRHS.add(s);
						}
						
						Rule newRule = new Rule(newName, newRHS);
						rules.add(newRule);
						rulesByLhs.addValue(newName, newRule);
					}
				}
				
				newNonTerminals.add(newName);
			}
			nextSymbols = nextNextSymbols;
			level++;
		}
		
		// Filter dead-end symbols
		nonTerminals = newNonTerminals;
		HashSet<String> deadEndSymbols = new HashSet<String>();
		// First find nonTerminals with no accompanying rules
		for (String s : nonTerminals) 
			if (!rulesByLhs.containsKey(s))
				deadEndSymbols.add(s);
		
		// Now for each nonTerminal, if its rules contain only
		// deadEnd symbols, add it too. Do so iteratively until no change
		int num = 0;
		while (num != deadEndSymbols.size()) {
			num = deadEndSymbols.size();
			for (String nonTerminal : nonTerminals) {
				if (deadEndSymbols.contains(nonTerminal))
					continue;
				
				boolean useful = false;
				for (Rule rule : rulesByLhs.getAll(nonTerminal)) {
					for (String s : rule.getRHS())
						if (!deadEndSymbols.contains(s)) {
							useful = true;
							break;
						}
					
					if (useful) break;
				}
				
				if (!useful)
					deadEndSymbols.add(nonTerminal);
			}
		}
		
		// remove all dead end symbols
		for (String deadEnd : deadEndSymbols) {
			nonTerminals.remove(deadEnd);
			rules.removeAll(rulesByLhs.getAll(deadEnd));
			rulesByLhs.removeKey(deadEnd);
		}
	}
	
	// gives equal probability to each rule
	public void finalizeRuleProbabilities() {
		for (String nonTerminal : nonTerminals) {
			HashSet<Rule> rules = rulesByLhs.getAll(nonTerminal);
			for (Rule rule : rules)
			{
				//for evedy non terminal that has probabilities - leave them
				if (rule.getProb() != 0.0)
					break;
				System.out.println("distributing unifromly non terminal " + nonTerminal);
				rule.prob = 1.0/rules.size();
			}
		}
	}
	
	public void clear() {
		rules.clear();
		rulesByLhs.clear();
		nonTerminals.clear();
	}
	
	public void printCFG (String filename)
	{
		try {
			FileOutputStream fos = new FileOutputStream(Finals.WORKING_DIR + filename);
			PrintWriter pw = new PrintWriter(fos);
		
			for (String nonTerminal : nonTerminals) {
				pw.println(nonTerminal);
				pw.println("-----------------");
			
				for (Rule r : rulesByLhs.getAll(nonTerminal))
					pw.println(r);
				
				pw.println();
			}
		
			pw.close();
			fos.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public String getStartSymbol()
	{
		return _startSymbol;
	}
	
	public void setStartSymbol(String startSymbol)
	{
		_startSymbol = startSymbol;
	}
	/*public static void addPattern(Pattern Patt,int RuleRemoveID)
	{
		rules.add(RuleRemoveID,new Rule(Patt.getLabel(), (Patt.getPath().toStringVector())));
	}
	
	public static void addPattern(Pattern Patt)
	{
		rules.add(new Rule(Patt.getLabel(), (Patt.getPath().toStringVector())));
	}*/

	
	/*public static void addEC(EquivalenceClass EC)
	{
		String ECLable = EC.getLabel();
		Iterator NodeIter=EC.getNodes().iterator();
		while (NodeIter.hasNext())
		{
			Path ECPath=new Path();
			ECPath.addNode((Node)NodeIter.next());
			rules.add(new Rule(ECLable,ECPath.toStringVector()));
		}
	}*/
	
	public Vector<Rule> getRules()
	{
		return rules;
	}
	
	public Rule getRule(int index) {
		return rules.get(index);
	}
	
	public boolean isNonTerminal(String node)
	{
		/*for (int i = 0; i < rules.size(); i++)
		{
			Rule currRule = (Rule) rules.get(i);
			if (currRule.getLHS().equals(node))
				return true;
		}
		return false;*/
		
		return nonTerminals.contains(node);
	}

	static Random generator;
	
	public void generateSentenceFiles(String file1, String file2, int size1, int size2, int recursionDepth) {
		try {
			FileOutputStream fos = new FileOutputStream(Finals.WORKING_DIR + file1);
			PrintWriter pw = new PrintWriter(fos);
			Vector<String> sentences = new Vector<String>();
			
			generator = new Random();
			int i = 0;
			while (i<size1) {
				String sentence = generateSentence(recursionDepth);
				if (sentence != null && !sentences.contains(sentence)) {
					pw.println("* " + sentence + " #");
					i++;
					sentences.add(sentence);
				}
			}
			
			pw.close();
			fos.close();
			
			fos = new FileOutputStream(Finals.WORKING_DIR + file2);
			pw = new PrintWriter(fos);
			i = 0;
			while (i<size2) {
				String sentence = generateSentence(recursionDepth);
				if (sentence != null && !sentences.contains(sentence)) {
					pw.println("* " + sentence + " #");
					i++;
					sentences.add(sentence);
				}
			}
			
			pw.close();
			fos.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public void generateSentenceFile(String filename, int numSentences, int recursionDepth, boolean disallowIdenticals) {
		try {
			FileOutputStream fos = new FileOutputStream(Finals.WORKING_DIR + filename);
			PrintWriter pw = new PrintWriter(fos);
			Vector<String> sentences = new Vector<String>();
			
			generator = new Random();
			int i = 0;
			while (i<numSentences) {
				String sentence = generateSentence(recursionDepth);
				if (sentence != null && (!disallowIdenticals || !sentences.contains(sentence))) {
					pw.println("* " + sentence + " #");
					i++;
					sentences.add(sentence);
				}
			}
			
			pw.close();
			fos.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	// Assumes there's a _startSymbol rule
	public String generateSentence(int recursionDepth) {
		//generator = new Random();
		String sentence = generateSentence(recursionDepth, _startSymbol);
		
		if (sentence == null)
			return null;
		
		StringTokenizer st = new StringTokenizer(sentence, " ");
		StringBuffer res = new StringBuffer();
		
		while (st.hasMoreTokens()) {
			res.append(st.nextToken());
			if (st.hasMoreTokens())
				res.append(" ");
		}	
		
		return res.toString();
	}
	
	private String generateSentence(int recursionDepth, String symbol) {
		if (recursionDepth == 0)
			return null;
		
		if (!isNonTerminal(symbol))
			return symbol;
		
		Vector<Rule> applicableRules = new Vector<Rule>();
		StringBuffer res = new StringBuffer();
		
		for (Rule rule : rules)
			if (rule.getLHS().equals(symbol))
				applicableRules.add(rule);
		
		Rule rule = applicableRules.get(generator.nextInt(applicableRules.size()));
		
		for (String subsymbol : rule.getRHS()) {
			String subsentence = generateSentence(recursionDepth-1, subsymbol);
			if (subsentence==null)
				return null;
			
			res.append(subsentence);
			res.append(" ");
		}
		
		return res.toString();
	}
	
	private static class SymbolWithDepth {
		public String symbol;
		public int depth;
		
		public SymbolWithDepth(String symbol, int depth) {
			this.symbol = symbol;
			this.depth = depth;
		}
	}
	
	private static class StringVectorComparator implements Comparator {
		public int compare(Object o1, Object o2) {
			Vector<String> v1 = (Vector<String>)o1;
			Vector<String> v2 = (Vector<String>)o2;
			
			int len = Math.min(v1.size(), v2.size());
			for (int i=0; i<len; i++) { 
				int comparison = v1.get(i).compareTo(v2.get(i));
				if (comparison != 0)
					return comparison;
			}
			
			return v1.size() - v2.size();
		}
	}
	
	public void save(String filename)
	{
		try
		{
			FileOutputStream fos = new FileOutputStream(Finals.WORKING_DIR + filename);
			PrintWriter pw = new PrintWriter(fos);
			for (Rule rule : rules) {
				pw.println(rule.toString());
			}
			pw.close();
			fos.close();
		} catch (Exception e)
		{
			e.printStackTrace();
		}

	}
}