package probParser;

import java.util.*;
import java.io.*;

import helpers.*;
import main.Finals;
import model.TreeNode;

public class CYK {
	CFG cfg;
	DoubleMap<String>[][] table;
	HashMap<String,Rule>[][] ruleTable;
	HashMap<String,Integer>[][] indexTable;
	
	public CYK(CFG cfg) {
		this.cfg = cfg;
	}
	
	public double parse(String sentence) {
		Vector<String> v = new Vector<String>(); 
		StringTokenizer st = new StringTokenizer(sentence);
		
		while (st.hasMoreTokens()) {
			String token = st.nextToken();
			if (token.equals("*") || token.equals("#"))
				continue;
			
			v.add(token);
		}
		return parse(v);
	}
	
	public TreeNode singleBestParse(String sentence) {
		Vector<String> v = new Vector<String>(); 
		StringTokenizer st = new StringTokenizer(sentence);
		
		while (st.hasMoreTokens()) {
			String token = st.nextToken();
			if (token.equals("*") || token.equals("#"))
				continue;
			
			v.add(token);
		}
		return singleBestParse(v);
	}
	
	public double parse(Vector<String> sentence) {
		
		table = new DoubleMap[sentence.size()][sentence.size()];
		
		for (int i=0; i<sentence.size(); i++) {
			table[0][i] = new DoubleMap<String>();
			HashSet<Rule> rules = cfg.rulesByTerminalRHS.getAll(sentence.get(i));
			for (Rule rule : rules)
				table[0][i].incexp(rule.LHS, Math.log(rule.prob));
		}
		
		for (int i=1; i<sentence.size(); i++) {
			for (int j=0; j<sentence.size()-i; j++) {
				table[i][j] = new DoubleMap<String>();
				
				for (int k=i-1; k >= 0; k--) {
					DoubleMap<String> rhs0 = table[k][j];
					DoubleMap<String> rhs1 = table[i-1-k][j+k+1];
				
					for (String s : rhs0.keySet()) {
						for (String s2 : rhs1.keySet()) {
							HashSet<Rule> rules = cfg.getRules(s, s2);
							
							if (rules == null)
								continue;
							
							for (Rule rule : rules) {
								double prob = rhs0.get(s) + rhs1.get(s2);
								prob += Math.log(rule.prob);
								table[i][j].incexp(rule.LHS, prob);
							}
						}
					}
				}
			}
		}
		
		//printChart(table);
		
		if (table[sentence.size()-1][0].containsKey(cfg.getStartSymbol()))
			return Math.exp(table[sentence.size()-1][0].get(cfg.getStartSymbol()));
		else
			return 0;
	}
	
	public TreeNode singleBestParse(Vector<String> sentence)
	{
		table = new DoubleMap[sentence.size()][sentence.size()];
		ruleTable = new HashMap[sentence.size()][sentence.size()];
		indexTable = new HashMap[sentence.size()][sentence.size()];
		
		for (int i=0; i<sentence.size(); i++) {
			table[0][i] = new DoubleMap<String>();
			ruleTable[0][i] = new HashMap<String,Rule>();
			indexTable[0][i] = new HashMap<String,Integer>();
			HashSet<Rule> rules = cfg.rulesByTerminalRHS.getAll(sentence.get(i));
			for (Rule rule : rules)
			{
				if(table[0][i].maxexp(rule.LHS, Math.log(rule.prob)))
					ruleTable[0][i].put(rule.getLHS(), rule);
				indexTable[0][i].put(rule.LHS, new Integer(0));
			}
		}
		
		for (int i=1; i<sentence.size(); i++) {

			for (int j=0; j<sentence.size()-i; j++) {
				table[i][j] = new DoubleMap<String>();
				ruleTable[i][j] = new HashMap<String,Rule>();
				indexTable[i][j] = new HashMap<String,Integer>();
				
				for (int k=i-1; k >= 0; k--) {
					DoubleMap<String> rhs0 = table[k][j];
					DoubleMap<String> rhs1 = table[i-1-k][j+k+1];
				
					for (String s : rhs0.keySet()) {
						for (String s2 : rhs1.keySet()) {
							HashSet<Rule> rules = cfg.getRules(s, s2);
							
							if (rules == null)
								continue;
							
							for (Rule rule : rules) {
								double prob = rhs0.get(s) + rhs1.get(s2);
								prob += Math.log(rule.prob);
								if(table[i][j].maxexp(rule.LHS, prob))
								{
									ruleTable[i][j].put(rule.getLHS(),rule);
									indexTable[i][j].put(rule.getLHS(), new Integer(k));
								}
							}
						}
					}
				}
			}
		}
		
		//printChart(table);
		
		if (table[sentence.size()-1][0].containsKey(cfg.getStartSymbol()))
			return new TreeNode(table,ruleTable,indexTable,sentence.size()-1,0,cfg.getStartSymbol());
		else
			return null;
		
		
	}
	
	public void parseCorpus(String unbracketedFile)
	{
		try
		{
			BufferedReader br = new BufferedReader(new FileReader(Finals.WORKING_DIR+unbracketedFile+Finals.EXTENSION));
			PrintWriter pw = new PrintWriter(new FileOutputStream(Finals.WORKING_DIR+unbracketedFile+"2"+Finals.EXTENSION));
			String line = null;
			while ((line = br.readLine())!=null)
			{
				TreeNode tree = singleBestParse(line);
				if(tree == null)
					System.out.print("");
				pw.println(tree.toPennString());
			}
			
			br.close();
			pw.close();
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}

	/*
	private void printChart(DoubleMap[][] chart) {
		try {
			FileOutputStream fos = new FileOutputStream(Finals.WORKING_DIR + "chart.txt");
			PrintWriter pw = new PrintWriter(fos);
			
			for (int i=0; i<chart.length; i++) {
				for (int j=0; j<chart[0].length; j++)
					pw.print(chart[i][j] + " | ");
					
				pw.println();
			}
			
			pw.close();
			fos.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	*/
}
