package probParser;

import java.util.*;

public class Rule 
{
	public Vector<String> RHS;
	public String LHS;
	//private boolean  IsTerminal;
	private int hash;
//	Changed For Merge
	public double prob;
	
	public CFG cfg;

	// Rule Constructor
	public Rule(String i_LHS, Vector<String> i_RHS)
	{
		RHS = i_RHS;
		LHS = i_LHS;
		prob = -1;

		hash = toShortString().hashCode();
	}
	
	public Rule(String i_LHS, Vector<String> i_RHS, double prob)
	{
		RHS = i_RHS;
		LHS = i_LHS;
		this.prob = prob;
		
		hash = toShortString().hashCode();
	}
	
	public Rule(String line) {
		StringTokenizer st = new StringTokenizer(line);
		String token = st.nextToken();
		if(!token.startsWith("["))
			LHS = token;
		else
		{
			StringBuffer sb = new StringBuffer();
			while(!token.endsWith("]"))
			{
				sb.append(token+ " ");
				token = st.nextToken();
			}
			sb.append(token);
			LHS = sb.toString();
		}
		st.nextToken(); // ->
		token = st.nextToken();
		
		RHS = new Vector<String>();
		while (!token.equals("(")) {
			if(!token.startsWith("["))
				RHS.add(token);
			else
			{
				StringBuffer sb = new StringBuffer();
				while(!token.endsWith("]"))
				{
					sb.append(token+ " ");
					token = st.nextToken();
				}
				sb.append(token);
				RHS.add(sb.toString());
			}
			token = st.nextToken();
		}
		
		prob = Double.valueOf(st.nextToken());
		hash = toShortString().hashCode();
	}
	
	public Vector<String> getRHS() 
	{
		return RHS;
	}
	
	public String getLHS()
	{
		return LHS;
	}
	
	public double getProb()
	{
		return prob;
	}
	 
	public String toShortString()
	 {
		 return LHS + " -> " + RHS.toString();		 
	 }
	
	 public String toString()
	 {
		 StringBuffer sb = new StringBuffer();
		 for (String str : RHS) {
			 sb.append(str);
			 sb.append(" ");
		 }
		 
		 return LHS + " -> " + sb.toString() + " ( " + prob + " )";			 
	 }
	 
	 public int hashCode() {
		 return hash;
	 }
	 
	 public boolean equals(Object o) {
		 Rule other = (Rule)o;
		 return RHS.equals(other.RHS) && LHS.equals(other.LHS);
	 }
	 
	 public boolean hasNonTerminals() {
		 for (String s : RHS)
			 if (cfg.isNonTerminal(s))
				 return true;
		 
		 return false;
	 }
	 
	 public Vector<String> nonTerminals() {
		 Vector<String> res = new Vector<String>();
		 for (String s : RHS)
			 if (cfg.isNonTerminal(s))
				 res.add(s);
		 
		 return res;
	 }
}
